# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

'''Interactive Principal Component Analysis of 2D images.

Computes Principal Components for a set of 2D images interactively until
stopped by user.

--- A note on Linear Algebra style terminology adopted here ---
Due to the specific nature of this PCA application, principal components are
referred to as 'eigenimages'. Each image from the original data set is
approximated by a linear combination of eigenimages, hence the coefficients
of this linear combination are referred to as 'coordinates' (of the approximate
image in the orthonormal basis of eigenimages) - in some other PCA software
these are referred to as 'reduced features data'.

The computed eigenimages and coordinates and the mean image are saved in the
file eigenimages.npz as three variables named eigim, coord and mean.

If you have docopt:
-------------------
Usage:
  icompute_eigenimages [--help | -h | options] <images>

Arguments:
  images  .npy file containing images as ndarray of dimensions (ni, ny, nx)

Options:
  -n <nim> , --nimgs=<nim>   number of images to use (< 0: all) [default: -1]
  -a <arch>, --arch=<arch>   architecture [default: cpu]

If you do not have docopt:
--------------------------
  icompute_eigenimages <images> <pca_err> [gpu]
'''

try:
    from docopt import docopt
    __version__ = '0.1.0'
    have_docopt = True
except:
    have_docopt = False

import math
import numpy
import pylab
import sys
import timeit

from raleigh.interfaces.pca import pca
from raleigh.core.solver import Options
from raleigh.interfaces.truncated_svd import UserStoppingCriteria

try:
    input = raw_input
except NameError:
    pass


def _norm(a, axis):
    return numpy.apply_along_axis(numpy.linalg.norm, axis, a)


class Probe:
    '''An object of this class will monitor the computation if passed via
       UserStoppingCriteria constructor to Options.stopping_criteria.
    '''
    def __init__(self, images):
        self.images = images
        self.shape = images.shape
        self.tol = 0 # start in interactive mode
        self.first = True # to give help hint at the start
        self.img_list = None # no images to show yet
        m = self.shape[0]
        ny = self.shape[1]
        nx = self.shape[2]
        self.nx = nx
        self.ny = ny
        # compute the norms of centered (shifted-to-zero-average) data samples
        # to be used for the relative truncation error computation
        n = nx*ny
        data = images.reshape((m, n))
        mean = numpy.mean(data, axis=0).reshape((1, n))
        ones = numpy.ones((m, 1), dtype=data.dtype)
        r = numpy.linalg.norm(mean)
        s = numpy.dot(data, mean.T)
        t = _norm(data, axis=1).reshape((m, 1))
        nrms_sqr = t*t - 2*s + r*r*ones
        self.nrms = numpy.sqrt(abs(nrms_sqr)).reshape((m,))

    def inspect(self, mean, sigma, left, right):
        '''Principal components computed by RALEIGH's method pca are the right
           singular vectors of the centered (shifted to zero mean) data matrix
           of shape samples x features, and coordinates of i-th data sample
           approximation are i-th components of the left singular vectors
           multiplied by the respective singular values.

           The method 'inspect' will be called by the RALEIGH's core eigensolver
           each time a bunch of singular vectors have converged.

           Parameters:
           -----------
           mean : 1D numpy array
                  mean row of the data matrix
           sigma: 1D numpy array
                  singular values of the data matrix computed so far
           left : 2D numpy array
                  columns are left singular vectors of the data matrix computed
                  so far
           right: 2D numpy array
                  columns are right singular vectors of the data matrix computed
                  so far

           In the implementation below, these data are, on user's request,
           used to visualize PCA approximations to images, helping the user to
           decide whether a sufficient number of eigenimages have been computed.
        '''
        # compute coordinates (reduced features data)
        s = numpy.reshape(sigma, (1, sigma.shape[0]))
        u = left*s
        # compute the errors
        proj = _norm(u, axis=1)
        nrms_sqr = self.nrms * self.nrms
        proj_sqr = proj*proj
        errs_sqr = nrms_sqr - proj_sqr # Pythagorean theorem
        err_fro = math.sqrt(numpy.sum(errs_sqr)/numpy.sum(nrms_sqr))
        i = sigma.shape[0] - 1 # latest singular value
        msg = 'sigma[%d] = %.1e*sigma[0], truncation error %.1e'
        msg = msg % (i, sigma[i]/sigma[0], err_fro)
        if self.tol > 0: # we are in non-intractive mode
            print(msg)
            if err_fro < self.tol:
                self.tol = 0 # back to interactive mode
            else:
                return False # more eigenimages needed
        while True: # process user's choices
            if self.first:
                print('answer h to the prompt below to get usage help')
                self.first = False
            ans = input(msg + ' h|q|s|t> ')
            vals = ans.split()
            nvals = len(vals)
            if nvals < 1:
                break
            if vals[0] == 'h':
                print('q               : to stop the computation')
                print('s im1 [im2 ...] : to show listed images')
                print('s               : to show previously selected images')
                print('t tolerance     : to continue in non-iteractive mode')
                print('                  until the truncation error falls')
                print('                  below tolerance')
                print('any other input : to compute few more eigenimages')
                continue
            if vals[0] == 'q':
                return True # we are done
            if vals[0] == 's':
                # visualize the current approximation to images
                nimgs = nvals - 1
                if nimgs < 1:
                    imgs = self.img_list
                    if imgs is None:
                        print('usage: s im1 [im2 ...]')
                        continue
                else:
                    imgs = vals[1:]
                nimgs = len(imgs)
                nim = left.shape[0]
                for i in range(nimgs):
                    im = int(imgs[i])
                    if im < 0 or im >= nim:
                        continue
                    image = self.images[im,:,:]
                    pylab.figure()
                    pylab.title('image %d' % im)
                    pylab.imshow(image, cmap='gray')
                    img = numpy.dot(u[im, :], right.T) + mean
                    pca_image = numpy.reshape(img, (ny, nx))
                    pylab.figure()
                    pylab.title('PCA approximation of the image %d' % im)
                    pylab.imshow(pca_image, cmap='gray')
                    pylab.show()
                self.img_list = imgs
                continue
            if vals[0] == 't':
                self.tol = float(vals[1])
            break
        return False # more eigenimages needed
        

if have_docopt:
    __version__ = '0.1.0'
    args = docopt(__doc__, version=__version__)
    file = args['<images>']
    ni = int(args['--nimgs'])
    arch = args['--arch']
else:
    narg = len(sys.argv)
    if narg < 2:
        print('Usage: icompute_eigenimages <images_file> [gpu]')
    file = sys.argv[1]
    arch = 'cpu' if narg < 4 else 'gpu!'
    ni = -1

numpy.random.seed(1) # make results reproducible

all_images = numpy.load(file)

m_all, ny, nx = all_images.shape
n = nx*ny

if ni < 0 or ni > m_all:
    ni = m_all

if ni < m_all:
    print('using first %d images only...' % ni)
    m = ni
    images = all_images[:m,:,:]
else:
    print('using all %d images...' % m_all)
    m = m_all
    images = all_images

# instruct the RALEIGH's core eigensolver to call the above implementation
# of Probe.inspect() each time a bunch of singular vectors have converged.
opt = Options()
probe = Probe(images)
opt.stopping_criteria = UserStoppingCriteria(images, shift=True, probe=probe)

images = numpy.reshape(images, (m, n))

mean, coord, eigim = pca(images, opt=opt, arch=arch, verb=1)

ncon = eigim.shape[0]
print('%d eigenimages computed' % ncon)

print('saving...')
mean = numpy.reshape(mean, (ny, nx))
eigim = numpy.reshape(eigim, (ncon, ny, nx))
numpy.savez('eigenimages', eigim=eigim, coord=coord, mean=mean)

print('done')



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Truncated SVD demo.

Usage:
  truncated_svd [--help | -h | <data> <rank>]

Arguments:
  data  numpy .npy file containing the matrix.
  rank  the number of singular values and vectors needed

"""

import numpy
import numpy.linalg as nla
from scipy.sparse.linalg import svds
import sys
import time

from raleigh.algebra import verbosity
verbosity.level = 2

from raleigh.core.solver import Options
from raleigh.interfaces.truncated_svd import truncated_svd


def _norm(a, axis):
    return numpy.apply_along_axis(nla.norm, axis, a)


narg = len(sys.argv)
if narg < 3 or sys.argv[1] == '-h' or sys.argv[1] == '--help':
    print('\nUsage:\n')
    print('python truncated_svd.py <data> <rank>')
    exit()
filename = sys.argv[1]
A = numpy.load(filename)
rank = int(sys.argv[2])
arch = 'cpu' if narg < 4 else 'gpu!'

numpy.random.seed(1) # make results reproducible

m = A.shape[0]
if len(A.shape) > 2:
    n = numpy.prod(A.shape[1:])
    A = numpy.reshape(A, (m, n))
else:
    n = A.shape[1]
dtype = A.dtype.type

print('\n--- solving with truncated_svd...\n')
start = time.time()
u, sigma, vt = truncated_svd(A, nsv=rank, arch=arch, verb=1)
stop = time.time()
time_tsvd = stop - start
print('\ntruncated_svd time: %.1e' % time_tsvd)
print('\n%d singular vectors computed' % sigma.shape[0])
D = A - numpy.dot(sigma*u, vt)
err = numpy.amax(_norm(D, axis=1))/numpy.amax(_norm(A, axis=1))
print('\ntruncation error %.1e' % err)

print('\n--- solving with svds...\n')
start = time.time()
u, sigma, vt = svds(A, k=rank)
stop = time.time()
time_svds = stop - start
print('\nsvds time: %.1e' % time_svds)
print('\n%d singular vectors computed' % sigma.shape[0])
D = A - numpy.dot(sigma*u, vt)
err = numpy.amax(_norm(D, axis=1))/numpy.amax(_norm(A, axis=1))
print('\ntruncation error %.1e' % err)

print('\ndone')



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

'''Discretized Laplace operators in 1, 2 and 3 dimensions"
'''

import numpy
import scipy.sparse as scs

def lap1d(n, a):
    h = a/(n + 1)
    d = numpy.ones((n,))/(h*h)
    return scs.spdiags([-d, 2*d, -d], [-1, 0, 1], n, n, format='csr')


def lap2d(nx, ny, ax, ay):
    L = lap1d(nx, ax)
    Ly = lap1d(ny, ay)
    L = scs.csr_matrix(scs.kron(scs.eye(ny), L) + scs.kron(Ly, scs.eye(nx)))
    return L


def lap3d(nx, ny, nz, ax, ay, az):
    L = lap2d(nx, ny, ax, ay)
    Lz = lap1d(nz, az)
    L = scs.csr_matrix(scs.kron(scs.eye(nz), L) + scs.kron(Lz, scs.eye(nx*ny)))
    return L





'''Example scripts.

eigenimages:
    Principal Component Analysis (PCA) demos for 2D images.

pca:
    PCA demos for general data.

buckling_evp:
    Computes buckling load factors and modes.

core_solver.py:
    Basic core solver usage demo.

sparse_evp.py:
    Computes several eigenvalues and eigenvectors of a real symmetric matrix.

sparse_evp_all_options.py:
    Same as sparse_evp.py but with more options.

truncated_svd.py:
    Computes truncated SVD.
'''



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Computes several leftmost eigenvalues alpha and corresponding eigenvectors v
of the buckling problem

    (K + alpha Ks)v = 0,

where K is the structure stiffness matrix, Ks is the stress stiffness matrix,
alpha is the buckling load factor and v is the buckling mode shape. Both K and
Ks are real symmetric, K is positive definite and Ks is generally indefinite.

The problem at hand is solved by raleigh's partial_hevp in buckling mode and,
for the sake of performance comparison, by scipy's eigsh in buckling mode.

In both solvers, the problem is transformed into

    (K + alpha0 Ks)v = beta K v,

for some positive alpha0 near alpha's of interest, and solved by inverse
iterations.

Mind that while placing alpha0 near the buckling load factors of interest
improves performance of both partial_svd anf eigsh, a value of alpha0
that is larger than the lowest buckling load factor may cause eigsh
to miss few lowest load factors (e.g. try one of the pairs of matrices
suggested below with alpha0 = 2.5 or higher).

A set of test matrices K and Ks of various sizes (from 8902 to 577643) in
Matrix Market format can be obtained from

https://www.dropbox.com/sh/vuloiyb3tjk3c33/AAAiyAcP6_dh7oC1UFdUiRp9a?dl=0

(soon to be also available on https://sparse.tamu.edu/).

--------------------------------------------------------------------------------
Requires MKL 10.3 or later (needs mkl_rt.dll on Windows, libmkl_rt.so on Linux).
The latest MKL can be istalled by pip install --user mkl.
--------------------------------------------------------------------------------

Usage:
    buckling_evp <path> <K> <Ks> <alpha0> <nev> [<tol>]

    <path>    data folder path
    <K>       name of the Matrix Market file containing stiffness matrix
    <Ks>      name of the Matrix Market file containing stress stiffness matrix
    <alpha0>  buckling factor shift: a positive value near the leftmost buckling
              load factors of interest (see NOTE below)
    <nev>     number of buckling modes wanted
    <tol>     error tolerance (optional, default: 1e-6)
"""

import numpy
from scipy.io import mmread
from scipy.sparse.linalg import eigsh
import sys
import time

from raleigh.algebra import verbosity
verbosity.level = 2

from raleigh.interfaces.partial_hevp import partial_hevp


narg = len(sys.argv)
if narg < 6 or sys.argv[1] == '-h' or sys.argv[1] == '--help':
    print('\nUsage:\n')
    print('python buckling_evp <path> <K> <Ks> <alpha0> <nev> [<tol>]')
    exit()
path = sys.argv[1]
if not path.endswith(('/', '\\')):
    path += '/'
K = path + sys.argv[2]
Ks = path + sys.argv[3]
alpha0 = float(sys.argv[4])
nev = int(sys.argv[5])
if narg > 6:
    tol = float(sys.argv[6])
else:
    tol = 1e-6

print('\n---reading the matrix K from %s...' % K)
A = mmread(K).tocsr()
print('\n---reading the matrix Ks from %s...' % Ks)
B = mmread(Ks).tocsr()

numpy.random.seed(1) # makes the results reproducible

print('\n---solving with raleigh partial_hevp...')
start = time.time()
vals, vecs, status = partial_hevp(A, B, buckling=True, sigma=-alpha0, \
                                  which=nev, tol=tol, verb=0)
stop = time.time()
raleigh_time = stop - start
if status != 0:
    print('partial_hevp execution status: %d' % status)
load_factors = -vals
print('buckling load factors computed:')
print(load_factors)
print('partial_hevp time: %.2e' % raleigh_time)

print('\n---solving with scipy eigsh...')
start = time.time()
vals, vecs = eigsh(A, nev, B, sigma=-alpha0, mode='buckling', which='LM', \
                   tol=tol)
stop = time.time()
eigsh_time = stop - start
load_factors = -vals[::-1]
print('buckling load factors computed:')
print(load_factors)
print('eigsh time: %.2e' % eigsh_time)



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

'''Principal Component Analysis of 2D images.

Computes Principal Components for a set of 2D images incrementally until
the error of PCA approximation falls below the required tolerance.

Note on Linear Algebra style terminology adopted here:
Due to a specific nature of this PCA application, principal components are
referred to as 'eigenimages'. Each image from the original data set is
approximated by a linear combination of eigenimages, hence we refer to the
coefficients of these linear combinations as 'coordinates' (of images in
the orthonormal basis of eigenimages) - in some other PCA software these
are referred to as 'reduced features data'.

If you have docopt:

Usage:
  compute_eigenimages [--help | -h | options] <images> <err_tol>

Arguments:
  images   .npy file containing images as ndarray of dimensions (ni, ny, nx)
  err_tol  image approximation error tolerance

Options:
  -n <nim> , --nimgs=<nim>   number of images to use (< 0: all) [default: -1]
  -a <arch>, --arch=<arch>   architecture [default: cpu]

If you do not have docopt:

  compute_eigenimages <images> <err_tol> [gpu]
'''

try:
    from docopt import docopt
    __version__ = '0.1.0'
    have_docopt = True
except:
    have_docopt = False

import math
import numpy
import pylab
import sys
import timeit

from raleigh.interfaces.pca import pca
from raleigh.core.solver import Options
from raleigh.interfaces.truncated_svd import UserStoppingCriteria


def _norm(a, axis):
    return numpy.apply_along_axis(numpy.linalg.norm, axis, a)


if have_docopt:
    __version__ = '0.1.0'
    args = docopt(__doc__, version=__version__)
    file = args['<images>']
    err_tol = float(args['<err_tol>'])
    ni = int(args['--nimgs'])
    arch = args['--arch']
else:
    narg = len(sys.argv)
    if narg < 3:
        print('Usage: compute_eigenimages <images> <err_tol> [gpu]')
    file = sys.argv[1]
    err_tol = float(sys.argv[2])
    arch = 'cpu' if narg < 4 else 'gpu!'
    ni = -1

numpy.random.seed(1) # make results reproducible

all_images = numpy.load(file)

m_all, ny, nx = all_images.shape
n = nx*ny

if ni < 0 or ni > m_all:
    ni = m_all

if ni < m_all:
    print('using first %d images only...' % ni)
    m = ni
    images = all_images[:m,:,:]
else:
    print('using all %d images...' % m_all)
    m = m_all
    images = all_images

images = numpy.reshape(images, (m, n))

start = timeit.default_timer()
mean, coord, eigim = pca(images, tol=err_tol, arch=arch, verb=1)
elapsed_time = timeit.default_timer() - start
ncon = eigim.shape[0]
print('%d eigenimages computed in %.2e sec' % (ncon, elapsed_time))

while True:
    i = int(input('image number (negative to exit): '))
    if i < 0 or i >= m:
        break
    pylab.figure()
    pylab.title('image %d' % i)
    img = images[i,:]
    image = numpy.reshape(img, (ny, nx))
    pylab.imshow(image, cmap='gray')
    img = numpy.dot(coord[i,:], eigim) + mean
    pca_image = numpy.reshape(img, (ny, nx))
    pylab.figure()
    pylab.title('PCA approximation of the image')
    pylab.imshow(pca_image, cmap='gray')
    pylab.show()

print('saving...')
mean = numpy.reshape(mean, (ny, nx))
eigim = numpy.reshape(eigim, (ncon, ny, nx))
numpy.savez('eigenimages', eigim=eigim, coord=coord, mean=mean)

print('done')



'''Example and utility scripts.

pca_simple.py:
    Computes a number of principal components of a dataset.
    If sklearn is installed, compares with sklearn.decomposition.PCA.

pca_smart.py:
    Computes PCA approximation of specified accuracy.
    If sklearn is installed, compares with restarted sklearn.decomposition.PCA.

pca_update.py:
    Performs PCA on a chunk of data, then addds more data and updates principal 
    components.

incremental_pca.py:
    Incremental Principal Component Analysis demo.
    Principal components are computed incrementally, processing a number 
    (batch size) of data samples at a time.
    For comparison, the same amount of principal components is computed also by 
    scikit-learn IncrementalPCA.

interactive_pca.py:
    Computes principal components of a dataset until stopped by the user.

generate_matrix.py:
   Generates a data matrix from randomly generated singular vectors and 
   singular values of user-controlled behaviour.
   If run as a script, the generated matrix is saved to data.npy, and singular
   values and vectors to svd.npz.
   Alternatively, from generate_matrix import generate, after which the call of
   generate(m, n, rank, dtype, scale, alpha, pca) will return the matrix,
   singular values and left and right singular vectors.

'''



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

# -*- coding: utf-8 -*-
'''Interactive PCA demo.

Computes principal components of a dataset until stopped by the user.

Usage: interactive_pca <data_file> [gpu]

data_file    : the name of the file containing data
gpu          : run raleigh pca on GPU if this argument is present
'''

import numpy
import os
import sys
import timeit

from raleigh.interfaces.pca import pca, pca_error
from raleigh.core.solver import Options
from raleigh.interfaces.truncated_svd import UserStoppingCriteria


narg = len(sys.argv)
if narg < 2:
    usage = 'Usage: interactive_pca <data_file> [gpu]'
    raise SystemExit(usage)
data = numpy.load(sys.argv[1])
m = data.shape[0]
n = data.shape[1]
if len(data.shape) > 2:
    n = numpy.prod(data.shape[1:])
    data = numpy.reshape(data, (m, n))
arch = 'cpu' if narg < 3 else 'gpu!'

numpy.random.seed(1) # make results reproducible

opt = Options()
opt.stopping_criteria = UserStoppingCriteria(data, shift=True)
mean, trans, comps = pca(data, opt=opt, arch=arch)
ncomp = comps.shape[0]
print('%d principal components computed' % ncomp)
em, ef = pca_error(data, mean, trans, comps)
print('PCA error: max 2-norm %.0e, Frobenius norm %.0e' % (em, ef))

print('done')



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Computes several eigenvalues and eigenvectors of a real symmetric matrix.

--------------------------------------------------------------------------------
Requires MKL 10.3 or later (needs mkl_rt.dll on Windows, libmkl_rt.so on Linux).
Latest MKL can be installed by pip install --user mkl
--------------------------------------------------------------------------------

Visit https://sparse.tamu.edu/ to download matrices in Matrix Market format
to test on (recommended group: DNVS).

Usage:
    sparse_evp <matrix> <nev> [<sigma>, [<tol>]]

    <matrix>  the name of the Matrix Market file containing problem matrix or
              parameters of 3D discretized Laplacian (mesh sizes nx, ny, nz
              and domain sizes ax, ay, az)
    <sigma>   shift (default: 0)
    <nev>     number of eigenvalues wanted nearest to the shift
    <tol>     error tolerance (default: 1e-6)
"""

import numpy
from scipy.io import mmread
from scipy.sparse.linalg import eigsh
import scipy.sparse as scs
import sys
import time

from raleigh.algebra import verbosity
verbosity.level = 2

try:
    from raleigh.interfaces.partial_hevp import partial_hevp
except:
    print('This script requires MKL 10.3 or later, sorry!')
    exit()


narg = len(sys.argv)
if narg < 3 or sys.argv[1] == '-h' or sys.argv[1] == '--help':
    print('\nUsage:\n')
    print('python sparse_evp <matrix> <nev> [<sigma>, [<tol>]]')
    exit()
matrix = sys.argv[1]
nev = int(sys.argv[2])
if narg > 3:
    sigma = float(sys.argv[3])
else:
    sigma = 0.0
if narg > 4:
    tol = float(sys.argv[4])
else:
    tol = 1e-6

i = matrix.find('.mtx')
if i < 0: # the file contains 3D Laplacian parameters
    f = open(matrix, 'r')
    data = f.read()
    par = data.split()
    nx = int(par[0])
    ny = int(par[1])
    nz = int(par[2])
    ax = float(par[3])
    ay = float(par[4])
    az = float(par[5])
    from raleigh.examples.laplace import lap3d
    print('\n---generating discretized 3D Laplacian matrix...')
    A = lap3d(nx, ny, nz, ax, ay, az)
else: # the file contains the matrix in Matrix Market format
    print('\n---reading the matrix from %s...' % matrix)
    A = mmread(matrix).tocsr()

numpy.random.seed(1) # makes the results reproducible

print('\n---solving with raleigh partial_hevp...')
start = time.time()
vals, vecs, status = partial_hevp(A, sigma=sigma, which=nev, tol=tol, verb=-1)
stop = time.time()
raleigh_time = stop - start
if status != 0:
    print('partial_hevp execution status: %d' % status)
print('converged eigenvalues are:')
print(vals)
print('raleigh time: %.2e' % raleigh_time)

print('\n---solving with scipy eigsh...')
start = time.time()
vals, vecs = scs.linalg.eigsh(A, nev, sigma=sigma, which='LM', tol=tol)
stop = time.time()
eigsh_time = stop - start
print(vals)
print('eigsh time: %.2e' % eigsh_time)
